/*
    Copyright 2009 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/stat.h>
#include "des.h"
#include "rwbits.h"
#include "show_dump2.h"

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;
typedef uint64_t    u64;



#define VER         "0.1"
//#define DEBUG       // hexdump of key, iv and data



extern void tiger(u8 *, u64, u8 *);
void hex2byte(u8 *in, u8 *out);
void quick_showhex(u8 *msg, u8 *data, int datalen);
int bits2str(u8 *out, int outsz, int bits, u8 *buff, int pos);
void get_secretsauce(u8 *nickname, u8 *cdkey, u8 *sauce);
int dwcrypto_bits(void *dest, u8 *data, int bits, int pos, u8 *key, u8 *iv);
int dwcryptonet(u8 *secretsauce, u8 *data, int datalen);
void dwcryptonet_decrypt(u8 *key, u8 *iv, u8 *data, int datalen);
u8 *fdload(u8 *fname, int *fsize);
void std_err(void);



static  u8  dwcrypto_key_1[24]  = "";



int main(int argc, char *argv[]) {
    int     datalen,
            plen;
    u8      secretsauce[100],   // 8 for secretsauce or 48 for dwcrypto_key_1
            *nick,
            *cdkey,
            *fname,
            *data,
            *p,
            *limit;

    fputs("\n"
        "DemonWare auth network decrypter "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 3) {
        printf("\n"
            "Usage: %s <dump_file> <secretsauce/dwcrypto_key_1>\n"
            "\n"
            "Note: for a easier usage dump ALL the content of the connections performed to\n"
            "      the demonware.net server in one file so the dwcrypto_key_1 will be\n"
            "      used automatically without reinserting it manually\n"
            "Note: if you don't know your secretsauce just put your nicknamd and your cdkey\n"
            "      example: dwcryptonet file.dat mynickname AAAABBBBCCCCDDDDEEEE\n"
            //"Usage: %s <dump_file> <nickname> <cdkey>\n"
            "\n", argv[0]);
        exit(1);
    }

    fname = argv[1];
    strcpy(secretsauce, argv[2]);

    if(argc > 3) {
        nick  = argv[2];
        cdkey = argv[3];
        get_secretsauce(nick, cdkey, secretsauce);
        printf("- secretsauce %s\n", secretsauce);
    }

    data = fdload(fname, &datalen);

    p     = data;
    limit = data + datalen;
    while(p < limit) {
        plen = *(u32 *)p;   p += 4;
        p += dwcryptonet(secretsauce, p, plen);
    }

    free(data);
    return(0);
}



void hex2byte(u8 *in, u8 *out) {
    int     t;

    while(in[0] && in[1]) {
        sscanf(in, "%02x", &t);
        *out++ = t;
        in += 2;
    }
}



void quick_showhex(u8 *msg, u8 *data, int datalen) {
    int     i;

    printf("%s", msg);
    for(i = 0; i < datalen; i++) printf("%02x", data[i]);
    printf("\n");
}



int bits2str(u8 *out, int outsz, int bits, u8 *buff, int pos) {
    u8      *o;

    if(outsz < (bits >> 3)) return(0);
    for(o = out; bits > 0; o++) {
        *o = read_bits(8, buff, pos);
        pos  += 8;
        bits -= 8;
    }
    return(o - out);
}



void get_secretsauce(u8 *nickname, u8 *cdkey, u8 *sauce) {
    int     i,
            nicklen;
    u8      c;

    nicklen = strlen(nickname);
    for(i = 0; i < 8; i++) {
        if(i < 8) {
            c = nickname[i];
        } else {
            c = 0;
        }
        c ^= cdkey[i];
        if(c > 0x7a) {
            c = 0x7a;
        } else if(c < 0x30) {
            c = 0x30;
        }
        sauce[i] = c;
    }
    sauce[i] = 0;
}



int dwcrypto_bits(void *dest, u8 *data, int bits, int pos, u8 *key, u8 *iv) {
    u32     t;

    if(dest) {
        if(bits <= 32) {
            t = read_bits(bits, data, pos);
            printf(". %-4d bits: %08x\n", bits, t);
            *(u32 *)dest = t;
        } else {
            t = bits2str(dest, 0x400, bits, data, pos);
            if(key) dwcryptonet_decrypt(key, iv, dest, t);
            printf(". %-4d bits%s:\n", bits, key ? " (decrypted)" : "");
            show_dump(7, dest, t, stdout);
        }
    }
    return(pos + bits);
}



int dwcryptonet(u8 *secretsauce, u8 *data, int datalen) {
    u32     n;
    int     b   = 0;
    u8      buff[0x400],
            key[24],
            iv[24],
            type;

    if(datalen < 2) return(datalen);
    if((data[0] == 0xff) && (data[1] == 0xff)) return(4);

    if(strlen(secretsauce) == 48) {
        hex2byte(secretsauce, dwcrypto_key_1);
    } else {
        tiger(secretsauce, strlen(secretsauce), key);
    }

    if(data[0]) {
        printf("\n- type direct decrypt\n");
        tiger(data + 1, 4, iv);
        data    += 1 + 4;
        datalen -= 1 + 4;
        dwcryptonet_decrypt(dwcrypto_key_1, iv, data, datalen);
        printf(". %-4d bits%s:\n", datalen, " (decrypted)");
        show_dump(7, data, datalen, stdout);
        return(datalen + 1 + 4);
    }

    type = data[1];
    printf("\n- type 0x%02x\n", type);
    data += 2;

    //if(type is invalid) {
        //n = read_bits(1,  data, b);  b += 1;
        //n = read_bits(5,  data, b);  b += 5;
        //return;
    //}

    b = dwcrypto_bits(NULL, data, 1,  b, NULL, NULL);
    b = dwcrypto_bits(NULL, data, 32, b, NULL, NULL);

    switch(type) {
        /* CLIENT */
        case 0x07: {
            b = 0;  // needed because the previous 2 dwcrypto_bits are for server
            b = dwcrypto_bits(&n, data, 1,  b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 5,  b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 5,  b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            b = dwcrypto_bits(buff, data, 0x400, b, NULL, NULL);
            break;
        }
        case 0x10: {
            b = 0;  // needed because the previous 2 dwcrypto_bits are for server
            b = dwcrypto_bits(&n, data, 1,  b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 5,  b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 5,  b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            break;
        }
        case 0x0a:
        case 0x14: {
            b = 0;  // needed because the previous 2 dwcrypto_bits are for server
            b = dwcrypto_bits(&n, data, 1,  b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 5,  b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 5,  b, NULL, NULL);
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            b = dwcrypto_bits(buff, data, 0x40, b, NULL, NULL);
            break;
        }

        /* SERVER */
        case 0x0b: {
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            tiger((u8 *)&n, 4, iv);
            b = dwcrypto_bits(buff, data, 0x400, b, key,  iv);  // user ticket
            memcpy(dwcrypto_key_1, buff + 0x61, 24);            // in cod5 is also performed a tiger hash on this
            quick_showhex("- set dwcrypto_key_1: ", dwcrypto_key_1, 24);
            b = dwcrypto_bits(buff, data, 0x400, b, NULL, NULL);
            break;
        }
        case 0x0d: {
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            tiger((u8 *)&n, 4, iv);
            b = dwcrypto_bits(buff, data, 0x400, b, key,  iv);  // user ticket
            break;
        }
        case 0x0f: {
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            tiger((u8 *)&n, 4, iv);
            b = dwcrypto_bits(buff, data, 0x400, b, key,  iv);  // user ticket
            break;
        }
        case 0x11: {
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            tiger((u8 *)&n, 4, iv);
            b = dwcrypto_bits(buff, data, 0x400, b, NULL, NULL); // user ticket
            memcpy(dwcrypto_key_1, buff + 0x61, 24);
            quick_showhex("- set dwcrypto_key_1: ", dwcrypto_key_1, 24);
            b = dwcrypto_bits(buff, data, 0x400, b, NULL, NULL);
            break;
        }
        case 0x15: {
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            tiger((u8 *)&n, 4, iv);
            b = dwcrypto_bits(buff, data, 0x200, b, NULL, NULL);
            break;
        }
        case 0x19: {
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            tiger((u8 *)&n, 4, iv);
            b = dwcrypto_bits(buff, data, 0x400, b, key,  iv);   // user ticket
            b = dwcrypto_bits(buff, data, 0x400, b, NULL, NULL); // LSG ticket
            b = dwcrypto_bits(buff, data, 0x2b0, b, NULL, NULL); // LSG ticket
            break;
        }
        case 0x1b: {
            b = dwcrypto_bits(&n, data, 32, b, NULL, NULL);
            tiger((u8 *)&n, 4, iv);
            b = dwcrypto_bits(buff, data, 0x400, b, key,  iv);   // user ticket
            b = dwcrypto_bits(buff, data, 0x400, b, NULL, NULL); // LSG ticket
            b = dwcrypto_bits(buff, data, 0x400, b, NULL, NULL); // encrypted cdkey
            break;
        }
        default: {
            printf("- unknown type 0x%x\n", type);
            break;
        }
    }
    return(datalen);
}



void dwcryptonet_decrypt(u8 *key, u8 *iv, u8 *data, int datalen) {
    des3_context    ctx;

#ifdef DEBUG
    printf("\n- 3des key:\n");
    show_dump(2, key, 24, stdout);
    printf("\n- 3des IV:\n");
    show_dump(2, iv, 24, stdout);
    printf("\n- encrypted data:\n");
    show_dump(2, data, datalen, stdout);
#endif

    des3_set3key_dec(&ctx, key);
    des3_crypt_cbc(&ctx, DES_DECRYPT, datalen, iv, data, data);

#ifdef DEBUG
    printf("\n- decrypted data:\n");
    show_dump(2, data, datalen, stdout);
#endif
}



u8 *fdload(u8 *fname, int *fsize) {
    struct stat xstat;
    FILE    *fd;
    int     size;
    u8      *buff;

    printf("- load file %s\n", fname);
    fd = fopen(fname, "rb");
    if(!fd) std_err();
    fstat(fileno(fd), &xstat);
    size = xstat.st_size;
    buff = malloc(size);
    if(!buff) std_err();
    fread(buff, 1, size, fd);
    fclose(fd);
    *fsize = size;
    return(buff);
}



void std_err(void) {
    perror("\nError");
    exit(1);
}


